# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# Templates from: https://github.com/JamesALeedham/Sphinx-Autosummary-Recursion/tree/master

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('../..'))


# -- Project information -----------------------------------------------------

project = 'AttriBench'
copyright = '2023, Arne Gevaert, Axel-Jan Rousseau'
author = 'Arne Gevaert, Axel-Jan Rousseau'

# The full version, including alpha/beta/rc tags
release = '0.0.1'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'numpydoc',  # Numpy documentation style
    'sphinx.ext.autodoc',  # Automatic documentation of docstrings
    'sphinx.ext.autosummary',  # Automatic generation of summary tables
    'sphinx.ext.viewcode',  # Add links to source code
    'sphinx_autodoc_typehints',  # Automatically document type hints for cleaner signatures
]
autoclass_content = 'both'
autodoc_inherit_docstrings = True
set_type_checking_flag = True
autodoc_default_options = {
    'members': True,
    'inherited-members': True,
}
numpydoc_show_class_members = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = ['css/custom.css']